from easylink.pipeline_schema_constants import development, tests

ALLOWED_SCHEMA_PARAMS = {
    "development": development.SCHEMA_PARAMS,
}

TESTING_SCHEMA_PARAMS = {
    "integration": tests.SINGLE_STEP_SCHEMA_PARAMS,
    "combined_bad_topology": tests.BAD_COMBINED_TOPOLOGY_SCHEMA_PARAMS,
    "combined_bad_implementation_names": tests.BAD_COMBINED_TOPOLOGY_SCHEMA_PARAMS,
}



"""Unit tests for the various Step classes.

Notes
-----
These unit tests often instantiate a Step object using parameters that would not
actually pass validation in a real-world scenario (i.e. they do not conform to the pipeline
schema). This is intentional because it's easier to flex complexity and edge cases here
rather than try to get full coverage in the e2e tests. It also allows us to test for
future pipeline schema expansion/flexibility in a relative simple manner now.
"""

from typing import Any

import pytest
from layered_config_tree import LayeredConfigTree

from easylink.configuration import Config
from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.pipeline_schema_constants.development import NODES
from easylink.step import (
    ChoiceStep,
    HierarchicalStep,
    IOStep,
    LoopStep,
    ParallelStep,
    Step,
)
from easylink.utilities.validation_utils import validate_input_file_dummy

STEP_KEYS = {step.name: step for step in NODES}


@pytest.fixture
def basic_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_1",
        "input_slots": [
            InputSlot(
                "step_1_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_1_main_output")],
    }


def test_implementation_node_name(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    node_name = step.implementation_node_name
    assert node_name == "step_1_python_pandas"

    step.set_parent_step(Step(step_name="foo", name="bar"))
    node_name = step.implementation_node_name
    assert node_name == "bar_step_1_step_1_python_pandas"


@pytest.fixture
def io_step_params() -> dict[str, Any]:
    return {
        "step_name": "io",
        "input_slots": [InputSlot("result", None, validate_input_file_dummy)],
        "output_slots": [OutputSlot("file1")],
    }


def test_io_step_slots(io_step_params: dict[str, Any]) -> None:
    step = IOStep(**io_step_params)
    assert step.name == step.step_name == "io"
    assert step.input_slots == {
        "result": InputSlot("result", None, validate_input_file_dummy)
    }
    assert step.output_slots == {"file1": OutputSlot("file1")}


def test_io_get_implementation_graph(
    io_step_params: dict[str, Any], default_config: Config
) -> None:
    step = IOStep(**io_step_params)
    step._get_config(default_config["pipeline"]["steps"])
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["io"]
    assert list(subgraph.edges) == []


def test_basic_step_slots(basic_step_params: dict[str, Any]) -> None:
    step = Step(**basic_step_params)
    assert step.name == step.step_name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_basic_step_get_implementation_graph(
    basic_step_params: dict[str, Any], default_config: Config
) -> None:
    step = Step(**basic_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_1_python_pandas"]
    assert list(subgraph.edges) == []


@pytest.fixture
def hierarchical_step_params() -> dict[str, Any]:
    return {
        "step_name": "step_4",
        "input_slots": [
            InputSlot(
                "step_4_main_input",
                "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validate_input_file_dummy,
            )
        ],
        "output_slots": [OutputSlot("step_4_main_output")],
        "nodes": [
            Step(
                "step_4a",
                input_slots=[
                    InputSlot(
                        "step_4a_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4a_main_output")],
            ),
            Step(
                "step_4b",
                input_slots=[
                    InputSlot(
                        "step_4b_main_input",
                        "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validate_input_file_dummy,
                    )
                ],
                output_slots=[OutputSlot("step_4b_main_output")],
            ),
        ],
        "edges": [
            EdgeParams("step_4a", "step_4b", "step_4a_main_output", "step_4b_main_input")
        ],
        "input_slot_mappings": [
            InputSlotMapping("step_4_main_input", "step_4a", "step_4a_main_input")
        ],
        "output_slot_mappings": [
            OutputSlotMapping("step_4_main_output", "step_4b", "step_4b_main_output")
        ],
    }


def test_hierarchical_step_slots(hierarchical_step_params: dict[str, Any]) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    assert step.name == "step_4"
    assert step.input_slots == {
        "step_4_main_input": InputSlot(
            "step_4_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_4_main_output": OutputSlot("step_4_main_output")}


def test_hierarchical_step_get_implementation_graph(
    hierarchical_step_params: dict[str, Any]
) -> None:
    step = HierarchicalStep(**hierarchical_step_params)
    pipeline_params = LayeredConfigTree(
        {"step_4": {"implementation": {"name": "step_4_python_pandas", "configuration": {}}}}
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_4_python_pandas"]
    assert list(subgraph.edges) == []

    # Test get_implementation_graph for substeps
    pipeline_params = LayeredConfigTree(
        {
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {
                            "name": "step_4a_python_pandas",
                            "configuration": {},
                        }
                    },
                    "step_4b": {
                        "implementation": {
                            "name": "step_4b_python_pandas",
                            "configuration": {},
                        }
                    },
                },
            },
        }
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_4a_python_pandas",
        "step_4b_python_pandas",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.fixture
def loop_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_3",
            input_slots=[
                InputSlot(
                    name="step_3_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
                InputSlot(
                    name="step_3_secondary_input",
                    env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_3_main_output")],
            nodes=[
                Step(
                    step_name="step_3a",
                    input_slots=[
                        InputSlot(
                            name="step_3a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3a_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3a_main_output")],
                ),
                Step(
                    step_name="step_3b",
                    input_slots=[
                        InputSlot(
                            name="step_3b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                        InputSlot(
                            name="step_3b_secondary_input",
                            env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_3b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_3a",
                    target_node="step_3b",
                    output_slot="step_3a_main_output",
                    input_slot="step_3b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_3_main_input",
                    child_node="step_3a",
                    child_slot="step_3a_main_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3a",
                    child_slot="step_3a_secondary_input",
                ),
                InputSlotMapping(
                    parent_slot="step_3_secondary_input",
                    child_node="step_3b",
                    child_slot="step_3b_secondary_input",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_3_main_output",
                    child_node="step_3b",
                    child_slot="step_3b_main_output",
                )
            ],
        ),
        "self_edges": [
            EdgeParams(
                source_node="step_3",
                target_node="step_3",
                output_slot="step_3_main_output",
                input_slot="step_3_main_input",
            )
        ],
    }


def test_loop_step_slots(loop_step_params: dict[str, Any]) -> None:
    step = LoopStep(**loop_step_params)
    assert step.name == step.step_name == "step_3"
    assert isinstance(step, LoopStep)
    assert step.input_slots == {
        "step_3_main_input": InputSlot(
            "step_3_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "step_3_secondary_input": InputSlot(
            "step_3_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {"step_3_main_output": OutputSlot("step_3_main_output")}
    assert isinstance(step.template_step, HierarchicalStep)
    assert step.template_step.name == step.name
    assert step.template_step.input_slots == step.input_slots
    assert step.template_step.output_slots == step.output_slots
    assert step.self_edges == [
        EdgeParams(step.name, step.name, "step_3_main_output", "step_3_main_input")
    ]


def test_loop_get_implementation_graph(
    mocker, loop_step_params: dict[str, Any], default_config: Config
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = LoopStep(**loop_step_params)
    step.set_configuration_state(default_config["pipeline"]["steps"], {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_3_python_pandas"]
    assert list(subgraph.edges) == []

    pipeline_params = LayeredConfigTree(
        {
            "step_3": {
                "iterate": [
                    {
                        "implementation": {
                            "name": "step_3_python_pandas",
                            "configuration": {},
                        },
                    },
                    {
                        "substeps": {
                            "step_3a": {
                                "implementation": {
                                    "name": "step_3a_python_pandas",
                                    "configuration": {},
                                },
                            },
                            "step_3b": {
                                "implementation": {
                                    "name": "step_3b_python_pandas",
                                    "configuration": {},
                                },
                            },
                        },
                    },
                ],
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_3_loop_1_step_3_python_pandas",
        "step_3_loop_2_step_3a_step_3a_python_pandas",
        "step_3_loop_2_step_3b_step_3b_python_pandas",
    ]
    expected_edges = [
        (
            "step_3_loop_1_step_3_python_pandas",
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3a_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_3_loop_2_step_3a_step_3a_python_pandas",
            "step_3_loop_2_step_3b_step_3b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_3b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_3a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.fixture
def parallel_step_params() -> dict[str, Any]:
    return {
        "template_step": HierarchicalStep(
            "step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
            nodes=[
                Step(
                    step_name="step_1a",
                    input_slots=[
                        InputSlot(
                            name="step_1a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1a_main_output")],
                ),
                Step(
                    step_name="step_1b",
                    input_slots=[
                        InputSlot(
                            name="step_1b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_1a",
                    target_node="step_1b",
                    output_slot="step_1a_main_output",
                    input_slot="step_1b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_1_main_input",
                    child_node="step_1a",
                    child_slot="step_1a_main_input",
                )
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_1_main_output",
                    child_node="step_1b",
                    child_slot="step_1b_main_output",
                )
            ],
        ),
    }


def test_parallel_step_slots(parallel_step_params: dict[str, Any]) -> None:
    step = ParallelStep(**parallel_step_params)
    assert step.name == "step_1"
    assert step.input_slots == {
        "step_1_main_input": InputSlot(
            "step_1_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        )
    }
    assert step.output_slots == {"step_1_main_output": OutputSlot("step_1_main_output")}


def test_parallel_step_get_implementation_graph(
    mocker, parallel_step_params: dict[str, Any]
) -> None:
    mocker.patch("easylink.implementation.Implementation._load_metadata")
    mocker.patch("easylink.implementation.Implementation.validate", return_value=[])
    step = ParallelStep(**parallel_step_params)
    pipeline_params = LayeredConfigTree(
        {
            "step_1": {
                "parallel": [
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_1",
                    },
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_2",
                    },
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                        "input_data_file": "input_file_3",
                    },
                ],
            },
        },
    )
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert set(subgraph.nodes) == {
        "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
        "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
        "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
    }
    expected_edges = [
        (
            "step_1_parallel_split_1_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_1_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_2_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_2_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_1_parallel_split_3_step_1a_step_1a_python_pandas",
            "step_1_parallel_split_3_step_1b_step_1b_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_1b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_1a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


@pytest.mark.parametrize("step_type", ["parallel", "loop"])
def test__duplicate_template_step(step_type):
    """Test against _duplicate_template_step.

    This is not an exhaustive test due to the complicated nature of testing
    equality of different attributes between two deep copies. For example,
    SomeClass.foo == SomeOtherClass.foo if foo is the same string even though
    they are different objects in memory. This is not the case for methods as
    well as other attribute types.
    """
    template_step = Step("step")
    template_step.set_configuration_state(
        LayeredConfigTree(
            {
                "step": {
                    "implementation": {
                        "name": "step_implementation",
                        "configuration": {},
                    },
                },
            },
        ),
        {},
        {},
    )
    if step_type == "loop":
        step = LoopStep(template_step)
        config_key = "iterate"
    else:  # parellel
        step = ParallelStep(template_step)
        config_key = "parallel"
    step.set_configuration_state(
        LayeredConfigTree(
            {
                "step": {
                    config_key: [
                        {
                            "implementation": {
                                "name": "step_implementation",
                                "configuration": {},
                            },
                        },
                    ],
                },
            },
        ),
        {},
        {},
    )

    duplicate_template_step = step._duplicate_template_step()

    special_handle_attrs = ["_configuration_state", "configuration_state", "step_graph"]
    attrs = [
        attr
        for attr in dir(step.template_step)
        if attr not in special_handle_attrs
        # dunders are too complicated to check
        and not attr.startswith("__")
        # methods are bound to each instance
        and not callable(getattr(step.template_step, attr))
    ]
    for attr in attrs:
        assert getattr(step.template_step, attr) == getattr(duplicate_template_step, attr)

    # Handle the special cases. Just check that they are of the same type and yet
    # not equal (implying that they are bound to different instances)
    for attr in special_handle_attrs:
        assert isinstance(
            getattr(step.template_step, attr),
            type(getattr(duplicate_template_step, attr)),
        )
        assert getattr(step.template_step, attr) != getattr(duplicate_template_step, attr)


@pytest.fixture
def choice_step_params() -> dict[str, Any]:
    return {
        "step_name": "choice_section",
        "input_slots": [
            InputSlot(
                name="choice_section_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
            InputSlot(
                name="choice_section_secondary_input",
                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            ),
        ],
        "output_slots": [OutputSlot("choice_section_main_output")],
        "choices": {
            "simple": {
                "nodes": [
                    HierarchicalStep(
                        step_name="step_4",
                        input_slots=[
                            InputSlot(
                                name="step_4_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                            InputSlot(
                                name="step_4_secondary_input",
                                env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_4_main_output")],
                        nodes=[
                            Step(
                                step_name="step_4a",
                                input_slots=[
                                    InputSlot(
                                        name="step_4a_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4a_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4a_main_output")],
                            ),
                            Step(
                                step_name="step_4b",
                                input_slots=[
                                    InputSlot(
                                        name="step_4b_main_input",
                                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                    InputSlot(
                                        name="step_4b_secondary_input",
                                        env_var="DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                                        validator=validate_input_file_dummy,
                                    ),
                                ],
                                output_slots=[OutputSlot("step_4b_main_output")],
                            ),
                        ],
                        edges=[
                            EdgeParams(
                                source_node="step_4a",
                                target_node="step_4b",
                                output_slot="step_4a_main_output",
                                input_slot="step_4b_main_input",
                            ),
                        ],
                        input_slot_mappings=[
                            InputSlotMapping(
                                parent_slot="step_4_main_input",
                                child_node="step_4a",
                                child_slot="step_4a_main_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4a",
                                child_slot="step_4a_secondary_input",
                            ),
                            InputSlotMapping(
                                parent_slot="step_4_secondary_input",
                                child_node="step_4b",
                                child_slot="step_4b_secondary_input",
                            ),
                        ],
                        output_slot_mappings=[
                            OutputSlotMapping(
                                parent_slot="step_4_main_output",
                                child_node="step_4b",
                                child_slot="step_4b_main_output",
                            ),
                        ],
                    ),
                ],
                "edges": [],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_4",
                        child_slot="step_4_main_input",
                    ),
                    InputSlotMapping(
                        parent_slot="choice_section_secondary_input",
                        child_node="step_4",
                        child_slot="step_4_secondary_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_4",
                        child_slot="step_4_main_output",
                    ),
                ],
            },
            "complex": {
                "nodes": [
                    Step(
                        step_name="step_5",
                        input_slots=[
                            InputSlot(
                                name="step_5_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_5_main_output")],
                    ),
                    # Add a more complicated (unsupported) loop step to ensure flexibility
                    LoopStep(
                        template_step=Step(
                            step_name="step_6",
                            input_slots=[
                                InputSlot(
                                    name="step_6_main_input",
                                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                    validator=validate_input_file_dummy,
                                ),
                            ],
                            output_slots=[OutputSlot("step_6_main_output")],
                        ),
                        self_edges=[
                            EdgeParams(
                                source_node="step_6",
                                target_node="step_6",
                                output_slot="step_6_main_output",
                                input_slot="step_6_main_input",
                            )
                        ],
                    ),
                ],
                "edges": [
                    EdgeParams(
                        source_node="step_5",
                        target_node="step_6",
                        output_slot="step_5_main_output",
                        input_slot="step_6_main_input",
                    )
                ],
                "input_slot_mappings": [
                    InputSlotMapping(
                        parent_slot="choice_section_main_input",
                        child_node="step_5",
                        child_slot="step_5_main_input",
                    ),
                ],
                "output_slot_mappings": [
                    OutputSlotMapping(
                        parent_slot="choice_section_main_output",
                        child_node="step_6",
                        child_slot="step_6_main_output",
                    ),
                ],
            },
        },
    }


def test_choice_step_slots(choice_step_params: dict[str, Any]) -> None:
    step = ChoiceStep(**choice_step_params)
    assert step.name == "choice_section"
    assert step.input_slots == {
        "choice_section_main_input": InputSlot(
            "choice_section_main_input",
            "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
        "choice_section_secondary_input": InputSlot(
            "choice_section_secondary_input",
            "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
            validate_input_file_dummy,
        ),
    }
    assert step.output_slots == {
        "choice_section_main_output": OutputSlot("choice_section_main_output")
    }


def test_simple_choice_step_get_implementation_graph(
    choice_step_params: dict[str, Any]
) -> None:
    step = ChoiceStep(**choice_step_params)

    # Test get_implementation_graph for single step (no substeps)
    pipeline_dict = {
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    # Need to validate in order to set the step graph an mappings prior to calling `set_configuration_state`
    step.validate_step(pipeline_dict["choice_section"], {}, {})
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == ["step_4_python_pandas"]
    assert list(subgraph.edges) == []

    # Test get_implementation_graph for a step with substeps
    pipeline_dict = {
        "choice_section": {
            "type": "simple",
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {
                            "name": "step_4a_python_pandas",
                        },
                    },
                    "step_4b": {
                        "implementation": {
                            "name": "step_4b_r",
                        },
                    },
                },
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_4a_python_pandas",
        "step_4b_r",
    ]
    expected_edges = [
        (
            "step_4a_python_pandas",
            "step_4b_r",
            {
                "input_slot": InputSlot(
                    "step_4b_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_4a_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)


def test_complex_choice_step_get_implementation_graph(
    choice_step_params: dict[str, Any]
) -> None:
    step = ChoiceStep(**choice_step_params)

    pipeline_dict = {
        "choice_section": {
            "type": "complex",
            "step_5": {
                "implementation": {
                    "name": "step_5_python_pandas",
                },
            },
            "step_6": {
                "iterate": [
                    {
                        "implementation": {
                            "name": "step_6_python_pandas",
                        },
                    },
                    {
                        "implementation": {
                            "name": "step_6_python_pandas",
                        },
                    },
                ],
            },
        },
    }
    pipeline_params = LayeredConfigTree(pipeline_dict)
    # Need to validate in order to set the step graph an mappings prior to calling `set_configuration_state`
    step.validate_step(pipeline_dict["choice_section"], {}, {})
    step.set_configuration_state(pipeline_params, {}, {})
    subgraph = step.get_implementation_graph()
    assert list(subgraph.nodes) == [
        "step_5_python_pandas",
        "step_6_loop_1_step_6_python_pandas",
        "step_6_loop_2_step_6_python_pandas",
    ]
    expected_edges = [
        (
            "step_5_python_pandas",
            "step_6_loop_1_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_5_main_output"),
                "filepaths": None,
            },
        ),
        (
            "step_6_loop_1_step_6_python_pandas",
            "step_6_loop_2_step_6_python_pandas",
            {
                "input_slot": InputSlot(
                    "step_6_main_input",
                    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validate_input_file_dummy,
                ),
                "output_slot": OutputSlot("step_6_main_output"),
                "filepaths": None,
            },
        ),
    ]
    assert len(subgraph.edges) == len(expected_edges)
    for edge in expected_edges:
        assert edge in subgraph.edges(data=True)



import csv
from pathlib import Path

import pytest
import yaml

from easylink.configuration import Config, load_params_from_specification

ENV_CONFIG_DICT = {
    "minimum": {
        "computing_environment": "local",
        "container_engine": "undefined",
    },
    "with_spark_and_slurm": {
        "computing_environment": "slurm",
        "container_engine": "singularity",
        "slurm": {
            "account": "some-account",
            "partition": "some-partition",
        },
        "implementation_resources": {
            "memory": 42,
            "cpus": 42,
            "time_limit": 42,
        },
        "spark": {
            "workers": {
                "num_workers": 42,
                "cpus_per_node": 42,
                "mem_per_node": 42,
                "time_limit": 42,
            },
        },
    },
}

PIPELINE_CONFIG_DICT = {
    "good": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "spark": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pyspark",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pyspark",
            },
        },
    },
    "out_of_order": {
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_step": {
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
    },
    "bad_step": {
        "foo": {  # Not a supported step
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_implementations": {
        "step_1": {
            "foo": "bar",  # Missing implementation key
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_implementation_name": {
        "step_1": {
            "implementation": {
                "foo": "bar",  # Missing name key
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "bad_implementation": {
        "step_1": {
            "implementation": {
                "name": "foo",  # Not a supported implementation
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_loop_nodes": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "iterate": [],
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "bad_loop_formatting": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "iterate": {"implementation": "step_3_python_pandas"},
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_substeps": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "substeps": {
                    "step_4a": {
                        "implementation": {},  # missing name key
                    },
                    # missing step_4b
                },
            },
        },
    },
    "wrong_parallel_split_keys": {
        "step_1": {
            "parallel": [
                {
                    "implementation": {
                        "name": "step_1_python_pandas",
                    },
                    "input_data_file": "foo",
                },
            ]
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "bad_combined_implementations": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "combined_implementation_key": "foo",
        },
        "choice_section": {
            "type": "simple",
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "missing_type_key": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            # missing 'type' key
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "bad_type_key": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "foo",  # Not a supported 'type'
            "step_4": {
                "implementation": {
                    "name": "step_4_python_pandas",
                },
            },
        },
    },
    "type_config_mismatch": {
        "step_1": {
            "implementation": {
                "name": "step_1_python_pandas",
            },
        },
        "step_2": {
            "implementation": {
                "name": "step_2_python_pandas",
            },
        },
        "step_3": {
            "implementation": {
                "name": "step_3_python_pandas",
            },
        },
        "choice_section": {
            "type": "simple",
            "step_5": {
                "implementation": {
                    "name": "step_5_python_pandas",
                },
            },
            "step_6": {
                "implementation": {
                    "name": "step_6_python_pandas",
                },
            },
        },
    },
}

INPUT_DATA_FORMAT_DICT = {
    "correct_cols": [["foo", "bar", "counter"], [1, 2, 3]],
    "wrong_cols": [["wrong", "column", "names"], [1, 2, 3]],
}

COMBINED_IMPLEMENTATION_CONFIGS = {
    "two_steps": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "implementation": {
                    "name": "step_2_python_pandas",
                },
            },
            "step_3": {
                "combined_implementation_key": "step_3_4",
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
    "with_iteration": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "implementation": {
                    "name": "step_2_python_pandas",
                },
            },
            "step_3": {
                "iterate": [
                    {
                        "implementation": {
                            "name": "step_3_python_pandas",
                        }
                    },
                    {
                        "combined_implementation_key": "step_3_4",
                    },
                ]
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
    "with_iteration_cycle": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "implementation": {
                    "name": "step_2_python_pandas",
                },
            },
            "step_3": {
                "iterate": [
                    {
                        "combined_implementation_key": "step_3_4",
                    },
                    {
                        "implementation": {
                            "name": "step_3_python_pandas",
                        }
                    },
                ]
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
    "with_parallel": {
        "steps": {
            "step_1": {
                "parallel": [
                    {
                        "input_data_file": "file1",
                        "implementation": {"name": "step_1_python_pandas"},
                    },
                    {
                        "input_data_file": "file2",
                        "implementation": {"name": "step_1_python_pandas"},
                    },
                    {
                        "input_data_file": "file2",
                        "combined_implementation_key": "steps_1_and_2_combined",
                    },
                ]
            },
            "step_2": {
                "combined_implementation_key": "steps_1_and_2_combined",
            },
            "step_3": {
                "implementation": {
                    "name": "step_3_python_pandas",
                }
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "implementation": {
                        "name": "step_4_python_pandas",
                    },
                },
            },
        },
        "combined_implementations": {
            "steps_1_and_2_combined": {
                "name": "step_1_and_step_2_combined_python_pandas",
            },
        },
    },
    "with_extra_node": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "combined_implementation_key": "step_3_4",
            },
            "step_3": {
                "combined_implementation_key": "step_3_4",
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
    "with_missing_node": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "implementation": {
                    "name": "step_2_python_pandas",
                },
            },
            "step_3": {
                "implementation": {
                    "name": "step_3_python_pandas",
                },
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
    "combined_bad_topology": {
        "steps": {
            "step_1": {
                "iterate": [
                    {
                        "substeps": {
                            "step_1a": {
                                "implementation": {
                                    "name": "step_1a_python_pandas",
                                },
                            },
                            "step_1b": {
                                "combined_implementation_key": "step_1a_1b",
                            },
                        },
                    },
                    {
                        "substeps": {
                            "step_1a": {
                                "combined_implementation_key": "step_1a_1b",
                            },
                            "step_1b": {
                                "implementation": {
                                    "name": "step_1b_python_pandas",
                                },
                            },
                        },
                    },
                ],
            },
        },
        "combined_implementations": {
            "step_1a_1b": {
                "name": "step_1a_and_step_1b_combined_python_pandas",
            },
        },
    },
    "combined_bad_implementation_names": {
        "steps": {
            "step_1": {
                "implementation": {
                    "name": "step_1_python_pandas",
                },
            },
            "step_2": {
                "implementation": {
                    "name": "step_2_python_pandas",
                },
                "combined_implementation_key": "step_3_4",  # incorrect key
            },
            "step_3": {
                "implementation": {
                    "name": "step_3_python_pandas",
                },
            },
            "choice_section": {
                "type": "simple",
                "step_4": {
                    "combined_implementation_key": "step_3_4",
                },
            },
        },
        "combined_implementations": {
            "step_3_4": {
                "name": "step_3_and_step_4_combined_python_pandas",
            },
        },
    },
}


def _write_csv(filepath: str, rows: list) -> None:
    with open(filepath, "w") as file:
        writer = csv.writer(file)
        writer.writerows(rows)


@pytest.fixture(scope="session")
def test_dir(tmpdir_factory) -> str:
    """Set up a persistent test directory with some of the specification files"""
    tmp_path = tmpdir_factory.getbasetemp()

    # good pipeline.yaml
    with open(f"{str(tmp_path)}/pipeline.yaml", "w") as file:
        yaml.dump({"steps": PIPELINE_CONFIG_DICT["good"]}, file, sort_keys=False)

    # dummy environment.yaml
    with open(f"{str(tmp_path)}/environment.yaml", "w") as file:
        yaml.dump(ENV_CONFIG_DICT["minimum"], file, sort_keys=False)
    with open(f"{str(tmp_path)}/spark_environment.yaml", "w") as file:
        yaml.dump(ENV_CONFIG_DICT["with_spark_and_slurm"], file, sort_keys=False)

    # input files
    input_dir1 = tmp_path.mkdir("input_data1")
    input_dir2 = tmp_path.mkdir("input_data2")
    for input_dir in [input_dir1, input_dir2]:
        for base_file in ["file1", "file2"]:
            # good input files
            _write_csv(input_dir / f"{base_file}.csv", INPUT_DATA_FORMAT_DICT["correct_cols"])
            # bad input files
            _write_csv(
                input_dir / f"broken_{base_file}.csv",
                INPUT_DATA_FORMAT_DICT["wrong_cols"],
            )
            # files with wrong extensions
            _write_csv(
                input_dir / f"{base_file}.oops",
                INPUT_DATA_FORMAT_DICT["correct_cols"],
            )

    # good input_data.yaml
    with open(f"{tmp_path}/input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "file1.csv"),
                "file2": str(input_dir2 / "file2.csv"),
            },
            file,
            sort_keys=False,
        )
    # input data is just a list (does not have keys)
    with open(f"{tmp_path}/input_data_list.yaml", "w") as file:
        yaml.dump(
            [
                str(input_dir1 / "file1.csv"),
                str(input_dir2 / "file2.csv"),
            ],
            file,
            sort_keys=False,
        )
    # missing input_data.yaml
    with open(f"{tmp_path}/missing_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "missing_file1.csv"),
            },
            file,
            sort_keys=False,
        )
    # input directs to files without sensible data
    with open(f"{tmp_path}/bad_columns_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "broken_file1.csv"),
                "file2": str(input_dir2 / "broken_file2.csv"),
            },
            file,
            sort_keys=False,
        )
    # incorrect file type
    with open(f"{tmp_path}/bad_type_input_data.yaml", "w") as file:
        yaml.dump(
            {
                "file1": str(input_dir1 / "file1.oops"),
                "file2": str(input_dir2 / "file2.oops"),
            },
            file,
            sort_keys=False,
        )

    # bad implementation
    bad_step_implementation_dir = tmp_path.join("steps/foo/implementations/bar/").ensure(
        dir=True
    )
    with open(f"{bad_step_implementation_dir}/metadata.yaml", "w") as file:
        yaml.dump(
            {
                "step": "foo",  # not a supported step
                "image": {
                    "path": "some/path/to/container/directory",
                    "filename": "bar",
                },
            },
            file,
            sort_keys=False,
        )

    return str(tmp_path)


@pytest.fixture()
def results_dir(test_dir) -> Path:
    return Path(f"{test_dir}/results_dir")


@pytest.fixture()
def default_config_paths(test_dir, results_dir) -> dict[str, Path]:
    return {
        "pipeline_specification": Path(f"{test_dir}/pipeline.yaml"),
        "input_data": Path(f"{test_dir}/input_data.yaml"),
        "computing_environment": Path(f"{test_dir}/environment.yaml"),
        "results_dir": results_dir,
    }


@pytest.fixture()
def default_config_params(default_config_paths) -> dict[str, Path]:
    return load_params_from_specification(**default_config_paths)


@pytest.fixture()
def default_config(default_config_params) -> Config:
    """A good/known Config object"""
    return Config(default_config_params)



"""
===============
Snakemake Rules
===============

We have chosen to use `Snakemake <https://snakemake.readthedocs.io/en/stable/>`_ 
as the EasyLink workflow manager. This module is responsible for generating the 
Snakemake rules to be run as well as writing them to the Snakefile.

Note we have adopted the Snakemake term "rule" to refer to a singular component 
in a Snakefile (i.e. in a Snakemake pipeline) that defines input files, output files,
and the command to run to create those output files. These rules are generated
dynamically as strings and appended to the Snakefile.
"""

import os
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass


class Rule(ABC):
    """An abstract class used to generate Snakemake rules."""

    def write_to_snakefile(self, snakefile_path) -> None:
        """Writes the rule to the Snakefile.

        Parameters
        ----------
        snakefile_path
            Path to the Snakefile to write the rule to.
        """
        with open(snakefile_path, "a") as f:
            f.write(self._build_rule())

    @abstractmethod
    def _build_rule(self) -> str:
        """Builds the snakemake rule to be written to the Snakefile.

        This is an abstract method and must be implemented by concrete instances.
        """
        pass


@dataclass
class TargetRule(Rule):
    """A :class:`Rule` that defines the final output of the pipeline.

    Snakemake will determine the directed acyclic graph (DAG) based on this target.
    """

    target_files: list[str]
    """List of final output filepaths."""
    validation: str
    """Name of file created by :class:`InputValidationRule`."""
    requires_spark: bool
    """Whether or not this rule requires a Spark environment to run."""

    def _build_rule(self) -> str:
        """Builds the Snakemake rule for the final output of the pipeline."""
        outputs = [os.path.basename(file_path) for file_path in self.target_files]
        rulestring = f"""
rule all:
    message: 'Grabbing final output'
    localrule: True   
    input:
        final_output={self.target_files},
        validation='{self.validation}',"""
        if self.requires_spark:
            rulestring += f"""
        term="spark_logs/spark_master_terminated.txt",
        master_log="spark_logs/spark_master_log.txt",
        worker_logs=gather.num_workers("spark_logs/spark_worker_log_{{scatteritem}}.txt",
        ),"""
        rulestring += f"""
    output: {outputs}
    run:
        import os
        for input_path, output_path in zip(input.final_output, output):
            os.symlink(input_path, output_path)"""
        return rulestring


@dataclass
class ImplementedRule(Rule):
    """A :class:`Rule` that defines the execution of an :class:`~easylink.implementation.Implementation`."""

    name: str
    """Name of the rule."""
    step_name: str
    """Name of the step this rule is implementing."""
    implementation_name: str
    """Name of the ``Implementation`` to build the rule for."""
    input_slots: dict[str, dict[str, str | list[str]]]
    """This ``Implementation's`` input slot attributes."""
    validations: list[str]
    """Names of files created by :class:`InputValidationRule`. These files are empty
    but used by Snakemake to build the graph edges of dependency on validation rules."""
    output: list[str]
    """Output data filepaths."""
    resources: dict | None
    """Computational resources used by executor (e.g. SLURM)."""
    envvars: dict
    """Environment variables to set."""
    diagnostics_dir: str
    """Directory for diagnostic files."""
    image_path: str
    """Path to the Singularity image to run."""
    script_cmd: str
    """Command to execute."""
    requires_spark: bool
    """Whether or not this ``Implementation`` requires a Spark environment."""

    def _build_rule(self) -> str:
        """Builds the Snakemake rule for this ``Implementation``."""
        return self._build_io() + self._build_resources() + self._build_shell_command()

    def _build_io(self) -> str:
        """Builds the input/output portion of the rule."""
        return (
            f"""
rule:
    name: "{self.name}"
    message: "Running {self.step_name} implementation: {self.implementation_name}" """
            + self._build_input()
            + f"""        
    output: {self.output}
    log: "{self.diagnostics_dir}/{self.name}-output.log"
    container: "{self.image_path}" """
        )

    def _build_input(self) -> str:
        input_str = f"""
    input:"""
        for slot_attrs in self.input_slots.values():
            input_str += f"""
        {slot_attrs["env_var"].lower()}={slot_attrs["filepaths"]},"""
        input_str += f"""
        validations={self.validations}, """
        if self.requires_spark:
            input_str += f"""
        master_trigger=gather.num_workers(rules.wait_for_spark_worker.output),
        master_url=rules.wait_for_spark_master.output,
            """
        return input_str

    def _build_resources(self) -> str:
        """Builds the resources portion of the rule."""
        if not self.resources:
            return ""
        return f"""
    resources:
        slurm_partition={self.resources['slurm_partition']},
        mem_mb={self.resources['mem_mb']},
        runtime={self.resources['runtime']},
        cpus_per_task={self.resources['cpus_per_task']},
        slurm_extra="--output '{self.diagnostics_dir}/{self.name}-slurm-%j.log'" """

    def _build_shell_command(self) -> str:
        """Builds the shell command portion of the rule."""
        shell_cmd = f"""
    shell:
        '''
        export DUMMY_CONTAINER_OUTPUT_PATHS={",".join(self.output)}
        export DUMMY_CONTAINER_DIAGNOSTICS_DIRECTORY={self.diagnostics_dir}"""
        for slot_attrs in self.input_slots.values():
            shell_cmd += f"""
        export {slot_attrs["env_var"]}={",".join(slot_attrs["filepaths"])}"""
        if self.requires_spark:
            shell_cmd += f"""
        read -r DUMMY_CONTAINER_SPARK_MASTER_URL < {{input.master_url}}
        export DUMMY_CONTAINER_SPARK_MASTER_URL"""
        for var_name, var_value in self.envvars.items():
            shell_cmd += f"""
        export {var_name}={var_value}"""
        # Log stdout/stderr to diagnostics directory
        shell_cmd += f"""
        {self.script_cmd} > {{log}} 2>&1
        '''"""

        return shell_cmd


@dataclass
class InputValidationRule(Rule):
    """A :class:`Rule` that validates input files.

    Each file coming into the pipeline via an :class:`~easylink.graph_components.InputSlot`
    must be validated against a specific validator function. This rule is responsible
    for running those validations as well as creating the (empty) validation output
    files that are used by Snakemake to build the graph edge from this rule to the
    next.
    """

    name: str
    """Name of the rule."""
    slot_name: str
    """Name of the ``InputSlot``."""
    input: list[str]
    """List of filepaths to validate."""
    output: str
    """Filepath of validation output. It must be used as an input for next rule."""
    validator: Callable
    """Callable that takes a filepath as input. Raises an error if invalid."""

    def _build_rule(self) -> str:
        return f"""
rule:
    name: "{self.name}_{self.slot_name}_validator"
    input: {self.input}
    output: touch("{self.output}")
    localrule: True         
    message: "Validating {self.name} input slot {self.slot_name}"
    run:
        for f in input:
            validation_utils.{self.validator.__name__}(f)"""



"""
========
Pipeline
========

This module is responsible for the ``Pipeline`` class, whose primary purpose
is to perform validations as well as generate the Snakefile to be used by 
`Snakemake <https://snakemake.readthedocs.io/en/stable/>`_ to execute the pipeline.

"""

from collections import defaultdict
from pathlib import Path

from loguru import logger

from easylink.configuration import Config
from easylink.pipeline_graph import PipelineGraph
from easylink.rule import ImplementedRule, InputValidationRule, TargetRule
from easylink.utilities.general_utils import exit_with_validation_error
from easylink.utilities.paths import SPARK_SNAKEFILE
from easylink.utilities.validation_utils import validate_input_file_dummy

IMPLEMENTATION_ERRORS_KEY = "IMPLEMENTATION ERRORS"


class Pipeline:
    """A convenience class for validations and Snakefile generation.

    Parameters
    ----------
    config
        The :class:`~easylink.configuration.Config` object.

    Attributes
    ----------
    config
        The :class:`~easylink.configuration.Config` object.
    pipeline_graph
        The :class:`~easylink.pipeline_graph.PipelineGraph` object.
    spark_is_required
        A boolean indicating whether the pipeline requires Spark.

    """

    def __init__(self, config: Config):
        self.config = config
        self.pipeline_graph = PipelineGraph(config)
        self.spark_is_required = self.pipeline_graph.spark_is_required()

        # TODO [MIC-4880]: refactor into validation object
        self._validate()

    @property
    def snakefile_path(self) -> Path:
        """The path to the dynamically-generated snakefile."""
        return self.config.results_dir / "Snakefile"

    def build_snakefile(self) -> Path:
        """Generates the Snakefile for this ``Pipeline``.

        This method dynamically builds the Snakefile by generating all necessary
        setup instructions (e.g. imports, configuration settings) as well as
        all rules for each :class:`~easylink.implementation.Implementation` in the
        pipeline and appending them to the Snakefile.

        Returns
        -------
            The path to the Snakefile.

        Notes
        -----
        We use the Snakemake term "rule" to refer to a singular component in a Snakefile
        (i.e. in a Snakemake pipeline) that defines input files, output files,
        and the command to run to create those output files. These rules are generated
        dynamically as strings and appended to the Snakefile.
        """
        if self.snakefile_path.is_file():
            logger.warning("Snakefile already exists, overwriting.")
            self.snakefile_path.unlink()
        self._write_imports()
        self._write_config()
        self._write_target_rules()
        if self.spark_is_required:
            self._write_spark_module()
        for node in self.pipeline_graph.implementation_nodes:
            self._write_implementation_rules(node)
        return self.snakefile_path

    ##################
    # Helper Methods #
    ##################

    def _validate(self) -> None:
        """Validates the pipeline.

        Raises
        ------
        SystemExit
            If any errors are found, they are batch-logged into a dictionary and
            the program exits with a non-zero code.
        """

        errors = {**self._validate_implementations()}

        if errors:
            exit_with_validation_error(errors)

    def _validate_implementations(self) -> dict:
        """Validates each individual :class:`~easylink.implementation.Implementation` instance.

        Returns
        -------
            A dictionary of ``Implementation`` validation errors.
        """
        errors = defaultdict(dict)
        for implementation in self.pipeline_graph.implementations:
            implementation_errors = implementation.validate()
            if implementation_errors:
                errors[IMPLEMENTATION_ERRORS_KEY][implementation.name] = implementation_errors
        return errors

    def _write_imports(self) -> None:
        """Writes the necessary imports to the Snakefile."""
        with open(self.snakefile_path, "a") as f:
            f.write("from easylink.utilities import validation_utils")

    def _write_config(self) -> None:
        """Writes configuration settings to the Snakefile.

        Notes
        -----
        This is currently only applicable for spark-dependent pipelines.
        """
        with open(self.snakefile_path, "a") as f:
            if self.spark_is_required:
                f.write(
                    f"\nscattergather:\n\tnum_workers={self.config.spark_resources['num_workers']},"
                )

    def _write_target_rules(self) -> None:
        """Writes the rule for the final output and its validation."""
        ## The "input" files to the result node/the target rule are the final output themselves.
        final_output, _ = self.pipeline_graph.get_io_filepaths("results")
        validator_file = str("input_validations/final_validator")
        # Snakemake resolves the DAG based on the first rule, so we put the target
        # before the validation
        target_rule = TargetRule(
            target_files=final_output,
            validation=validator_file,
            requires_spark=self.spark_is_required,
        )
        final_validation = InputValidationRule(
            name="results",
            slot_name="main_input",
            input=final_output,
            output=validator_file,
            validator=validate_input_file_dummy,
        )
        target_rule.write_to_snakefile(self.snakefile_path)
        final_validation.write_to_snakefile(self.snakefile_path)

    def _write_spark_module(self) -> None:
        """Inserts the ``easylink.utilities.spark.smk`` Snakemake module into the Snakefile."""
        slurm_resources = self.config.slurm_resources
        spark_resources = self.config.spark_resources
        with open(self.snakefile_path, "a") as f:
            module = f"""
module spark_cluster:
    snakefile: '{SPARK_SNAKEFILE}'
    config: config

use rule * from spark_cluster
use rule terminate_spark from spark_cluster with:
    input: rules.all.input.final_output"""
            if self.config.computing_environment == "slurm":
                module += f"""
use rule start_spark_master from spark_cluster with:
    resources:
        slurm_account={slurm_resources['slurm_account']},
        slurm_partition={slurm_resources['slurm_partition']},
        mem_mb={spark_resources['slurm_mem_mb']},
        runtime={spark_resources['runtime']},
        cpus_per_task={spark_resources['cpus_per_task']},
        slurm_extra="--output 'spark_logs/start_spark_master-slurm-%j.log'"
use rule start_spark_worker from spark_cluster with:
    resources:
        slurm_account={slurm_resources['slurm_account']},
        slurm_partition={slurm_resources['slurm_partition']},
        mem_mb={spark_resources['slurm_mem_mb']},
        runtime={spark_resources['runtime']},
        cpus_per_task={spark_resources['cpus_per_task']},
        slurm_extra="--output 'spark_logs/start_spark_worker-slurm-%j.log'"
    params:
        terminate_file_name=rules.terminate_spark.output,
        user=os.environ["USER"],
        cores={spark_resources['cpus_per_task']},
        memory={spark_resources['mem_mb']}
                        """
            f.write(module)

    def _write_implementation_rules(self, node_name: str) -> None:
        """Writes the rules for each :class:`~easylink.implementation.Implementation`.

        Parameters
        ----------
        node_name
            The name of the ``Implementation`` to write the rule(s) for.
        """
        implementation = self.pipeline_graph.nodes[node_name]["implementation"]
        _input_files, output_files = self.pipeline_graph.get_io_filepaths(node_name)
        input_slots = self.pipeline_graph.get_input_slot_attributes(node_name)
        diagnostics_dir = Path("diagnostics") / node_name
        diagnostics_dir.mkdir(parents=True, exist_ok=True)
        resources = (
            self.config.slurm_resources
            if self.config.computing_environment == "slurm"
            else None
        )
        validation_files, validation_rules = self._get_validations(node_name, input_slots)
        implementation_rule = ImplementedRule(
            name=node_name,
            step_name=" and ".join(implementation.metadata_steps),
            implementation_name=implementation.name,
            input_slots=input_slots,
            validations=validation_files,
            output=output_files,
            resources=resources,
            envvars=implementation.environment_variables,
            diagnostics_dir=str(diagnostics_dir),
            image_path=implementation.singularity_image_path,
            script_cmd=implementation.script_cmd,
            requires_spark=implementation.requires_spark,
        )
        for validation_rule in validation_rules:
            validation_rule.write_to_snakefile(self.snakefile_path)
        implementation_rule.write_to_snakefile(self.snakefile_path)

    @staticmethod
    def _get_validations(
        node_name, input_slots
    ) -> tuple[list[str], list[InputValidationRule]]:
        """Gets the validation rule and its output filepath for each slot for a given node.

        Parameters
        ----------
        node_name
            The name of the ``Implementation`` to get validation data for.
        input_slots
            The input slot attributes for the given node.

        Returns
        -------
            A tuple of lists containing the validation output paths and rules.
        """
        validation_files = []
        validation_rules = []

        for input_slot_name, input_slot_attrs in input_slots.items():
            validation_file = f"input_validations/{node_name}/{input_slot_name}_validator"
            validation_files.append(validation_file)
            validation_rules.append(
                InputValidationRule(
                    name=node_name,
                    slot_name=input_slot_name,
                    input=input_slot_attrs["filepaths"],
                    output=validation_file,
                    validator=input_slot_attrs["validator"],
                )
            )
        return validation_files, validation_rules



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from easylink.rule import ImplementedRule, InputValidationRule, Rule, TargetRule

RULE_STRINGS = {
    "target_rule": "rule_strings/target_rule.txt",
    "implemented_rule_local": "rule_strings/implemented_rule_local.txt",
    "implemented_rule_slurm": "rule_strings/implemented_rule_slurm.txt",
    "validation_rule": "rule_strings/validation_rule.txt",
}


def test_rule_interface():
    class TestRule(Rule):
        def _build_rule(self) -> str:
            return "'Tis but a scratch!"

    rule = TestRule()
    with TemporaryDirectory() as snake_dir_path:
        snakefile_path = Path(snake_dir_path) / "Snakefile"
        rule.write_to_snakefile(snakefile_path)
        with open(snakefile_path, "r") as f:
            assert f.read() == "'Tis but a scratch!"


def test_target_rule_build_rule():
    rule = TargetRule(
        target_files=["int/foo", "int/bar"],
        validation="bar",
        requires_spark=False,
    )
    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["target_rule"]
    with open(file_path) as expected_file:
        expected = expected_file.read()
    rulestring = rule._build_rule()
    rulestring_lines = rulestring.split("\n")
    expected_lines = expected.split("\n")
    assert len(rulestring_lines) == len(expected_lines)
    for i, expected_line in enumerate(expected_lines):
        assert rulestring_lines[i].strip() == expected_line.strip()


@pytest.mark.parametrize("computing_environment", ["local", "slurm"])
def test_implemented_rule_build_rule(computing_environment):
    if computing_environment == "slurm":
        resources = {
            "slurm_partition": "'slurmpart'",
            "runtime": 1,
            "mem_mb": 5120,
            "cpus_per_task": 1337,
        }
    else:
        resources = None

    rule = ImplementedRule(
        name="foo_rule",
        step_name="foo_step",
        implementation_name="foo_imp",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": lambda x: None,
            },
            "secondary": {
                "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                "filepaths": ["bar"],
                "validator": lambda x: None,
            },
        },
        validations=["bar"],
        output=["baz"],
        resources=resources,
        envvars={"eggs": "coconut"},
        diagnostics_dir="spam",
        image_path="Multipolarity.sif",
        script_cmd="echo hello world",
        requires_spark=False,
    )

    file_path = (
        Path(os.path.dirname(__file__))
        / RULE_STRINGS[f"implemented_rule_{computing_environment}"]
    )
    with open(file_path) as expected_file:
        expected = expected_file.read()
    rulestring = rule._build_rule()
    rulestring_lines = rulestring.split("\n")
    expected_lines = expected.split("\n")
    assert len(rulestring_lines) == len(expected_lines)
    for i, expected_line in enumerate(expected_lines):
        assert rulestring_lines[i].strip() == expected_line.strip()


def bar():
    pass


def test_validation_rule_build_rule():
    rule = InputValidationRule(
        name="foo", slot_name="foo_input", input=["foo", "bar"], output="baz", validator=bar
    )
    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["validation_rule"]
    with open(file_path) as expected_file:
        expected = expected_file.read()
    rulestring = rule._build_rule()
    rulestring_lines = rulestring.split("\n")
    expected_lines = expected.split("\n")
    assert len(rulestring_lines) == len(expected_lines)
    for i, expected_line in enumerate(expected_lines):
        assert rulestring_lines[i].strip() == expected_line.strip()
